#define TINY_GSM_MODEM_ESP8266
#define SerialMon Serial
#define SerialAT Serial1

#include <Adafruit_CircuitPlayground.h>
#include <TinyGsmClient.h>
#include <PubSubClient.h>

const char MQTT_CLIENTID[] = __DATE__ __TIME__;

// WiFi connection credentials, if applicable
const char wifiSSID[] = "Dan&Leah";
const char wifiPass[] = "28031997";

// MQTT details
const char* broker = "192.168.0.100";
const char temperatureChallengeTopic[] = "temperatureChallenge";
const char lightChallengeTopic[] = "lightChallenge";
const char soundChallengeTopic[] = "soundChallenge";
const char wakeUpTopic[] = "wakeUp";
const char chooseColorsTopic[] = "chooseColors";
const char jumpChallengeTopic[] = "jumpChallenge";

// Gloabal Variables
bool choseColors = false;
int countColors = 0;
int* firstLed;
int* secondLed;
int* thirdLed;
int* fourthLed;
int* allLeds[4];
uint32_t lastReconnectAttempt = 0;

TinyGsm modem(SerialAT);
TinyGsmClient client(modem);
PubSubClient mqtt(client);


void setup() {
  CircuitPlayground.begin();
  SerialMon.begin(115200);
  delay(1000);
  SerialAT.begin(115200);
  modem.restart();
  if (!modem.networkConnect(wifiSSID, wifiPass)) {
    SerialMon.println("ap connect fail ");
    while (true);
  }
  if (!modem.waitForNetwork()) {
    SerialMon.println("network fail");
    while (true);
  }
  if (!modem.isNetworkConnected()) {
    SerialMon.println("Network disconnected");
    while (true);
  }
  mqtt.setServer(broker, 1883);
  mqtt.setKeepAlive(10000000000); // So that mqtt connection is kept alive for the entire process
  mqtt.setCallback(callback);
}

void loop() {
  while (!mqttCheckConn());
}

/******************* Led and RGB Control Section *******************/

// function for setting all rgb leds
void ledsRGB(int red, int green, int blue) {
  for (int i = 0; i < 10; i++) {
    CircuitPlayground.setPixelColor(i, red, green, blue);
  }
}

// function that sets led lights color to be like police
void policeLights() {
  ledsRGB(0, 0, 255);
  delay(200);
  ledsRGB(255, 0, 0);
  delay(150);
}

// turns payload from mqtt to RGB array
int* turnToRgbArray(byte *payload, int length) {
  String inputString = String((char*)payload);
  int index1 = inputString.indexOf(':');
  int index2 = inputString.indexOf(';');
  String red = inputString.substring(0, index1);
  String green = inputString.substring(index1 + 1, index2);
  String blue = inputString.substring(index2 + 1, length);
  int *rgbArray = new int[3];
  rgbArray[0] = red.toInt();
  rgbArray[1] = green.toInt();
  rgbArray[2] = blue.toInt();
  return rgbArray;
}

/******************* MQTT Functions Section *******************/

// connection to the mqtt server
boolean mqttConnect() {
  SerialMon.print("Connecting to ");
  SerialMon.print(broker);
  boolean status = mqtt.connect(MQTT_CLIENTID);
  if (status == false) {
    SerialMon.println(" fail");
    return false;
  }
  SerialMon.println(" success");
  mqtt.subscribe(wakeUpTopic);
  mqtt.subscribe(chooseColorsTopic);
  return mqtt.connected();
}

// checks connection to the mqtt server
bool mqttCheckConn() {
  if (!mqtt.connected()) {
    uint32_t t = millis();
    if (t - lastReconnectAttempt > 10000L) {
      lastReconnectAttempt = t;
      if (mqttConnect()) {
        lastReconnectAttempt = 0;
        return true;
      }
    }
    delay(100);
    return false;
  }
  mqtt.loop();
  // disconnects occur otherwise
  delay(1000);
}

// callback function is called whenever a msg enteres the queue (not by us) handles waking up (startFlow) and getting colors from red-node
void callback(char* topic, byte* payload, unsigned int length) {
  if (String(topic) == wakeUpTopic) {
    String payloadString = String((char*)payload, length);
    if (payloadString == "WAKE UP!!!") {
      startFlow();
    }
  } else if (String(topic) == chooseColorsTopic && !choseColors) {
    countColors++;
    switch (countColors) {
      case 1:
        firstLed = turnToRgbArray(payload, length);
        break;
      case 2:
        secondLed = turnToRgbArray(payload, length);
        break;
      case 3:
        thirdLed = turnToRgbArray(payload, length);
        break;
      case 4:
        fourthLed = turnToRgbArray(payload, length);
        choseColors = true;
        allLeds[0] = firstLed;
        allLeds[1] = secondLed;
        allLeds[2] = thirdLed;
        allLeds[3] = fourthLed;
        break;
    }
  }
}

// General function for publishing to mqtt
void publishToMqtt(const char* topicName, String mqttStr) {
  char mqttBuff[50];
  mqttStr.toCharArray(mqttBuff, mqttStr.length() + 1);
  mqtt.publish(topicName, mqttBuff);
}

/******************* Main Flow and CheckContinue Section *******************/

// Calles the main functions for the flow
void startFlow() {
  alarmAndHeat();
  checkContinue();
  lightsChallenge();
  checkContinue();
  soundChallenge();
  checkContinue();
  jumpChallenge();
}

// Checks if user is ready to continue by pressing both buttons
void checkContinue() {
  while (!(CircuitPlayground.leftButton() && CircuitPlayground.rightButton()));
}

/******************* Temperature Challange Section *******************/

// Activates the geating challenge and makes the alarm go off until user succeeds
void alarmAndHeat() {
  publishToMqtt(temperatureChallengeTopic, "0");
  while (getTemp() < 30) {
    CircuitPlayground.playTone(440, 50);
    delay(50);
    CircuitPlayground.playTone(1200, 50);
    policeLights();
  }
  ledsRGB(0, 255, 0);
}

// Gets temp from cpx and publishes to queue for red-node
float getTemp() {
  String tempStr;
  long lastPublished = 0;
  float ftemp;
  if (millis() - lastPublished > 100) {
    ftemp = CircuitPlayground.temperature();
    tempStr = String(ftemp);
    publishToMqtt(temperatureChallengeTopic, tempStr);
    lastPublished = millis();
  }
  return ftemp;
}

/******************* Light Challange Section *******************/

// Function that starts the light game, gets the colors from mqtt, and calls the challenge flow
void lightsChallenge() {
  bool lightsMissionComplete = false;
  ledsRGB(0, 0, 0);
  String hallway = "Go to the hallway for the next challenge!";
  publishToMqtt(lightChallengeTopic, hallway);
  while (!choseColors) {
    mqtt.loop();
  }
  delay(2000);
  while (!lightsMissionComplete) {
    lightsMissionComplete = lightChallengeFlow();
    //lightsMissionComplete = true;
    ledsRGB(0, 0, 0);
  }
  ledsRGB(0, 255, 0);
  String message = "Well Done!";
  publishToMqtt(lightChallengeTopic, message);
}

// Shows the current light sequance for challenge
void showLightSequance(int currentRound[5], int ledNums[], int loopLength) {
  for (int j = 0; j < loopLength; j++) {
    int red = allLeds[currentRound[j]][0];
    int green = allLeds[currentRound[j]][1];
    int blue = allLeds[currentRound[j]][2];
    CircuitPlayground.setPixelColor(ledNums[currentRound[j]], red, green, blue);
    delay(1000);
    CircuitPlayground.setPixelColor(ledNums[currentRound[j]], 0, 0, 0);
    delay(1000);
  }
}

// Catches touch sensors and validates that it corresponds to the led sequance
bool catchTouchSensors(int currentRound[5], int loopLength, int ledNums[]) {
  byte sensePins[] = {A4, A5, A2, A3};
  for (int k = 0; k < loopLength; k++) {
    bool touchCorrect = false;
    while (!touchCorrect) {
      for (int l = 0; l < sizeof(sensePins); l++) {
        int value = CircuitPlayground.readCap(sensePins[l]);
        if (value >= 1017) {
          if (currentRound[k] != l) {
            ledsRGB(255, 0, 0);
            delay(2000);
            return false;
          }
          int red = allLeds[l][0];
          int green = allLeds[l][1];
          int blue = allLeds[l][2];
          CircuitPlayground.setPixelColor(ledNums[l], red, green, blue);
          delay(1000);
          CircuitPlayground.setPixelColor(ledNums[l], 0, 0, 0);
          delay(1000);
          touchCorrect = true;
          break;
        }
      }
    }
  }
  return true;
}

// Manages the light challenge flow
bool lightChallengeFlow () {
  int ledNums[] = {0, 1, 8, 9};
  int rounds[][5] = {{1, 0, 1, 2, 3}, {2, 3, 1, 0, 1}, {2, 0 , 3, 1, 3}};
  int loopLength = 5;
  for (int roundNum = 0; roundNum < 3; roundNum++) {
    showLightSequance(rounds[roundNum], ledNums, loopLength);
    if (!catchTouchSensors(rounds[roundNum], loopLength, ledNums)) {
      return false;
    }
    ledsRGB(0, 255, 0);
    delay(300);
    ledsRGB(0, 0, 0);
  }
  return true;
}

/******************* Sound Challange Section *******************/

// Gets sound level from sensor and publishes to mqtt
float getSound() {
  String soundStr;
  long lastPublished = 0;
  float fsound;
  if (millis() - lastPublished > 100) {
    fsound = CircuitPlayground.mic.soundPressureLevel(10);
    fsound = map(fsound, 55, 100, 0, 100);
    soundStr = String(fsound);
    publishToMqtt(soundChallengeTopic, soundStr);
    lastPublished = millis();
  }
  return fsound;
}

// Manages the sound challange
void soundChallenge() {
  publishToMqtt(soundChallengeTopic, "0");
  while (getSound() < 100) {
    policeLights();
  }
  String message = "Well Done!";
  publishToMqtt(soundChallengeTopic, message);
  ledsRGB(0, 255, 0);
}

/******************* Star Jumps Section *******************/

// Manages the Star jumps challange
void jumpChallenge() {
  float xValue = CircuitPlayground.motionX();
  float yValue = CircuitPlayground.motionY();
  float zValue = CircuitPlayground.motionZ();
  int jumpsLeft = 20;
  String jump = String(jumpsLeft);
  publishToMqtt(jumpChallengeTopic, jump);
  while (jumpsLeft > 0) {
    xValue = CircuitPlayground.motionX();
    yValue = CircuitPlayground.motionY();
    zValue = CircuitPlayground.motionZ();
    if (xValue > 15 || yValue > 15 || zValue > 15) {
      jumpsLeft--;
      jump = String(jumpsLeft);
      CircuitPlayground.playTone(440, 50);
      publishToMqtt(jumpChallengeTopic, jump);
    }
  }
}
